/*
 * Copyright 2016 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.uru.algo;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.jku.risc.stout.uru.data.Hedge;
import at.jku.risc.stout.uru.data.TermNode;
import at.jku.risc.stout.uru.data.atom.HedgeVar;
import at.jku.risc.stout.uru.data.atom.TermAtom;
import at.jku.risc.stout.uru.data.atom.TermVar;
import at.jku.risc.stout.uru.data.atom.Variable;

public class UnifAlgorithm {
	private final UnifSystem system;
	private final long maxDerivationDepth;
	private boolean maxDepthReached = false;
	private final Set<Substitution> result = new HashSet<>();

	public UnifAlgorithm(UnifSystem system, long maxDerivationDepth) {
		this.system = system;
		this.maxDerivationDepth = maxDerivationDepth;
		while (!system.getProblem().isEmpty()
				&& system.getProblem().getFirst().getLeft()
						.equals(system.getProblem().getFirst().getRight()))
			system.getProblem().remove();
	}

	public Set<Substitution> unify(DebugLevel debugLevel, PrintStream debugOut,
			PrintStream debugErr, boolean justify) {
		// Projection
		final UnifSystemCollection systems = projection(system);
		if (debugLevel == DebugLevel.PROGRESS) {
			debugOut.println();
			debugOut.println("Initial system: " + system);
			debugOut.println("  ==>  Projection ==>");
			debugOut.println("  Generated systems: " + systems);
			debugOut.println();
		}

		while (!systems.isEmpty()) {
			UnifSystem sys = systems.remove();
			if (debugLevel == DebugLevel.PROGRESS)
				debugOut.println("Selected system: " + sys);
			while (!sys.getProblem().isEmpty()) {
				UnifEquation eq = sys.getProblem().remove();
				if (eq.getDerivationDepth() >= maxDerivationDepth) {
					// Max derivation depth reached ===>> BOT
					sys.getProblem().add(eq);
					maxDepthReached = true;
					error("Derivation depth " + maxDerivationDepth + " reached",
							sys, debugLevel, debugErr);
					break;
				}
				eq.incDerivationDepth();
				TermAtom leftA = eq.getLeft().getAtom();
				TermAtom rightA = eq.getRight().getAtom();
				Hedge leftH = eq.getLeft().getHedge();
				Hedge rightH = eq.getRight().getHedge();
				if (isClash(leftH, rightH)) {
					sys.getProblem().add(eq);
					error("Irreparable difference encountered", sys,
							debugLevel, debugErr);
					break;
				}
				if (!leftA.equals(rightA)) {
					if (rightA instanceof TermVar
							&& !(leftA instanceof TermVar)) {
						// Orient 1
						eq.swap();
						sys.getProblem().add(eq);
						debug("Orient 1", sys, debugLevel, debugOut, false);
					} else if (leftA instanceof TermVar
							&& rightA != TermAtom.nullAtom) {
						if (eq.getRight().occurs(leftA)) {
							// Individual Variable Occurrence Check
							sys.getProblem().add(eq);
							error("Individual Variable Occurrence Check", sys,
									debugLevel, debugErr);
							break;
						} else {
							// Solve
							sys.apply((TermVar) leftA, eq.getRight());
							debug("Solve", sys, debugLevel, debugOut, false);
						}
					} else {
						// No rule applicable ===>> BOT
						sys.getProblem().add(eq);
						error("Symbol Clash", sys, debugLevel, debugErr);
						break;
					}
				} else if (leftH.equals(rightH)) {
					// Trivial
					debug("Trivial", sys, debugLevel, debugOut, false);
				} else if (rightH.size() == 0 || leftH.size() == 0) {
					// No rule applicable ===>> BOT
					sys.getProblem().add(eq);
					error("Empty", sys, debugLevel, debugErr);
					break;
				} else if (leftH.get(0).getAtom() instanceof HedgeVar) {
					if (leftH.get(0).getAtom() == rightH.get(0).getAtom()) {
						// Sequence Variable Elimination 1
						leftH.getSequence().remove(0);
						rightH.getSequence().remove(0);
						sys.getProblem().add(eq);
						debug("Sequence Variable Elimination 1", sys,
								debugLevel, debugOut, false);
					} else {
						HedgeVar leftVar = (HedgeVar) leftH.get(0).getAtom();
						if (!rightH.get(0).occurs(leftVar)) {
							// Widening 1
							UnifSystem sysNew = sys.copy();
							UnifEquation eqNew = eq.copy();
							List<TermNode> leftSeq = eqNew.getLeft().getHedge()
									.getSequence();
							TermNode x = leftSeq.remove(0);
							Hedge widen = new Hedge();
							widen.add(eqNew.getRight().getHedge().getSequence()
									.remove(0)); // t
							widen.add(x);
							sysNew.getProblem().add(eqNew);
							sysNew.apply(leftVar, new TermNode(null, widen));
							leftSeq.add(0, x);
							systems.add(sysNew);
							debug("Widening 1", sysNew, debugLevel, debugOut,
									true);
							if (rightH.get(0).getAtom() instanceof HedgeVar) {
								// Widening 2
								sysNew = sys.copy();
								eqNew = eq.copy();
								List<TermNode> rightSeq = eqNew.getRight()
										.getHedge().getSequence();
								TermNode y = rightSeq.remove(0);
								widen = new Hedge();
								widen.add(eqNew.getLeft().getHedge()
										.getSequence().remove(0)); // x
								widen.add(y);
								sysNew.getProblem().add(eqNew);
								sysNew.apply((Variable) y.getAtom(),
										new TermNode(null, widen));
								rightSeq.add(0, y);
								systems.add(sysNew);
								debug("Widening 2", sysNew, debugLevel,
										debugOut, true);
							}
							// Sequence Variable Elimination 2
							leftH.getSequence().remove(0);
							TermNode t = rightH.getSequence().remove(0);
							sys.getProblem().add(eq);
							sys.apply(leftVar, t);
							debug("Sequence Variable Elimination 2", sys,
									debugLevel, debugOut, false);
						} else {
							// No rule applicable ===>> BOT
							sys.getProblem().add(eq);
							error("Sequence Variable Occurrence Check", sys,
									debugLevel, debugErr);
							break;
						}
					}
				} else if (rightH.get(0).getAtom() instanceof HedgeVar) {
					// Orient 2
					eq.swap();
					sys.getProblem().add(eq);
					debug("Orient 2", sys, debugLevel, debugOut, false);
				} else {
					int idx = idxHedgeVar(leftH, rightH);
					if (idx != -1) {
						// Partial Decomposition 1
						for (int i = 0; i < idx; i++) {
							UnifEquation eqDec = new UnifEquation(leftH.get(i),
									rightH.get(i));
							eqDec.setDerivationDepth(eq.getDerivationDepth());
							sys.getProblem().add(eqDec);
						}
						eq.setLeft(new TermNode(leftA, leftH.subHedge(idx,
								leftH.size())));
						eq.setRight(new TermNode(rightA, rightH.subHedge(idx,
								rightH.size())));
						sys.getProblem().add(eq);
						debug("Partial Decomposition 1", sys, debugLevel,
								debugOut, false);
					} else if (leftH.size() == rightH.size()) {
						// Total Decomposition
						for (int i = leftH.size() - 1; i >= 0; i--) {
							UnifEquation eqDec = new UnifEquation(leftH.get(i),
									rightH.get(i));
							eqDec.setDerivationDepth(eq.getDerivationDepth());
							sys.getProblem().add(eqDec);
						}
						if (leftA != TermAtom.nullAtom || leftH.size() != 1)
							debug("Total Decomposition", sys, debugLevel,
									debugOut, false); // Non-trivial
					} else {
						// No rule applicable ===>> BOT
						sys.getProblem().add(eq);
						error("Arity Disagreement", sys, debugLevel, debugOut);
						break;
					}
				}
			}
			// Empty problem set = solution found
			if (sys.getProblem().isEmpty()) {
				if (justify) {
					UnifProblem problemOrig = system.getProblem().copy();
					while (!problemOrig.isEmpty()) {
						UnifEquation eqOrig = problemOrig.remove();
						eqOrig.apply(sys.getSigma());
						if (!eqOrig.getLeft().equals(eqOrig.getRight())) {
							debugErr.println("ERROR: Justification for the following unifier failed!!!");
							debugErr.println("ERROR: Unifier: "
									+ sys.getSigma());
							return null;
						}
					}
					if (debugLevel == DebugLevel.PROGRESS) {
						debugOut.println("  Justification OK for unifier: "
								+ sys.getSigma());
						debugOut.println();
					}
				}
				result.add(sys.getSigma());
			}
		}
		if (maxDepthReached)
			debugOut.println("ATTENTION: Maximum derivation depth "
					+ maxDerivationDepth
					+ " has been reached. More unifiers might exist.");
		if (debugLevel != DebugLevel.SILENT) {
			if (result.isEmpty())
				debugOut.println("No unifier found.");
			else
				debugOut.println("Unifiers found: " + result);
		}

		return result;
	}

	private boolean isClash(Hedge leftH, Hedge rightH) {
		Map<TermAtom, Integer> lHeadV = leftH.headV();
		Map<TermAtom, Integer> rHeadV = rightH.headV();
		Set<TermAtom> lHeadVkeys = lHeadV.keySet();
		Set<TermAtom> rHeadVkeys = lHeadV.keySet();
		boolean isHedgeVar = false;
		for (TermAtom x : lHeadVkeys) {
			if (x instanceof HedgeVar) {
				isHedgeVar = true;
				break;
			}
		}
		if (!isHedgeVar) {
			for (TermAtom x : rHeadVkeys) {
				if (x instanceof HedgeVar) {
					isHedgeVar = true;
					break;
				}
			}
			if (!isHedgeVar)
				return false;
		}
		Map<TermAtom, Integer> lHeadF = leftH.headF();
		Map<TermAtom, Integer> rHeadF = rightH.headF();
		if (multiSetContainsAll(lHeadV, rHeadV))
			if (!multiSetContainsAll(rHeadF, lHeadF))
				return true;
		if (multiSetContainsAll(rHeadV, lHeadV))
			if (!multiSetContainsAll(lHeadF, rHeadF))
				return true;
		return false;
	}

	private boolean multiSetContainsAll(
			Map<TermAtom, Integer> checkContainment,
			Map<TermAtom, Integer> elementsToCheck) {
		for (Entry<TermAtom, Integer> element : elementsToCheck.entrySet()) {
			Integer cnt = checkContainment.get(element.getKey());
			if (cnt == null || cnt < element.getValue())
				return false;
		}
		return true;
	}

	private void debug(String msg, UnifSystem sys, DebugLevel debugLevel,
			PrintStream out, boolean newBranch) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  ==>  " + msg + " ==>");
			if (newBranch)
				out.println("  New branch generated!");
		}
		debug(debugLevel, out, sys);
	}

	private void error(String msg, UnifSystem sys, DebugLevel debugLevel,
			PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  ==>  " + msg);
			out.println();
		}
	}

	private void debug(DebugLevel debugLevel, PrintStream out, UnifSystem sys) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("  System: " + sys);
			out.println();
		}
	}

	private int idxHedgeVar(Hedge leftH, Hedge rightH) {
		for (int n = Math.min(leftH.size(), rightH.size()), i = 0; i < n; i++)
			if (leftH.get(i).getAtom() instanceof HedgeVar
					|| rightH.get(i).getAtom() instanceof HedgeVar)
				return i;
		return -1;
	}

	private UnifSystemCollection projection(UnifSystem sys) {
		UnifSystemCollection systems = new UnifSystemCollection();
		systems.add(sys.copy());
		Set<HedgeVar> varSet = sys.getProblem().collectHedgeVars();
		HedgeVar[] vars = new HedgeVar[varSet.size()];
		int n = 1 << varSet.toArray(vars).length;
		for (int counter = 1; counter < n; counter++) {
			UnifSystem sysProj = sys.copy();
			systems.add(sysProj);
			for (int j = 0; j < vars.length; j++)
				if ((counter & (1 << j)) != 0)
					sysProj.eliminateSeqVariable(vars[j]);
		}
		return systems;
	}

	public boolean isMaxDepthReached() {
		return maxDepthReached;
	}

	public UnifSystem getInitialSystem() {
		return system;
	}

	public long getMaxDerivationDepth() {
		return maxDerivationDepth;
	}

	public Set<Substitution> getResult() {
		return result;
	}
}
